/*
  Pixy2 responds to the following library commands sent via the I2C interface
  For the colour connected components mode we include .ccc.

  pixy.ccc.blocks[]         - an array of block data, if they exist
  pixy.changeProg(cmd)      - change the primary mode of Pixy2
    cmd = color_connected_components – In this program mode the Pixy2 is detecting object “signatures” or Color Codes
    cmd = line_tracking              – In this program mode the Pixy2 is detecting lines, intersections, and barcodes
    cmd = pan_and_tilt_demo          – A demonstration program for the optional pan and tilt assembly
    cmd = video                      – In this mode the Pixy2 functions like a standard video camera
  pixy.ccc.getBlocks()      - instructs the Pixy2 camera to look for objects
  pixy.getVersion()         - returns the version number as a int8_t
  pixy.init()               - should be called before your program communicates with Pixy.
  pixy.ccc.numBlock         - a variable containing the number of blocks identified
  pixy.setCameraBrightness(brightness) - sets the relative exposure level of Pixy2's image sensor. 
  pixy.setLamp(U,L)         - U refers to the two white LEDS at the top, L refers to the RGB LED
  pixy.setLED(r,g,b)        - sets the RGB colour of the front LED
  pixy.setServos(S0,S1)     - set S0 and S1 servo values as uint16_t
  pixy.getResolution()      - loads the the frameWidth and frameHeight variables with the corresponding values
  pixy.getFPS()             - returns the frame rate between 2 - 62 depending on light levels

*/

// --------------------------------------------------------------------------------

void Pixy_Check() {
  // checks I2C for existance of Pixy2, normally called every 40ms if no response
  if (PixyDel > 0) {PixyDel--; return;}
  
  Wire.beginTransmission(0x54);
  I2C_Error = Wire.endTransmission();
  if (I2C_Error == 0) {
    I2C_Pixy = true;
//    Serial.println("Pixy2 Detected after " + String(millis()) + "ms");
    pixy.init();
    Pixy_ServoReset();
  } else {
    I2C_Pixy = false;
//    Serial.println("Pixy2 Not Found!");
  }
  PixyDel = 25; // put a 1 sec delay on next event

}

// --------------------------------------------------------------------------------

int8_t Pixy_GetBlocks() {
  // grab blocks from the camera
  // int8_t getBlocks(bool wait [optional], uint8_t sigmap [optional], uint8_t maxBlocks [optional])
  return pixy.ccc.getBlocks(false,255,BlockAsk); // polling function for blocks, does not wait for valid blocks
}

// --------------------------------------------------------------------------------

void LoadPixyBlocks() {
  // load array variables with valid blocks
  // note Pixy2 might return a number for blocks, but they are invalid
  // so we load them into an internal array for global use
//  t0[0] = micros();
  BlockCnt = pixy.ccc.numBlocks;
  if (BlockCnt > 0) {
    int16_t zValid = pixy.ccc.blocks[0].m_x;
    if (zValid > 0) {
      // looks ok so read in data
      for (int zb = 0;zb < BlockCnt;zb++) {
        // take each block in turn, remap and draw it
        Pixy_mX[zb] = pixy.ccc.blocks[zb].m_x;
        Pixy_mY[zb] = pixy.ccc.blocks[zb].m_y;
        Pixy_mW[zb] = pixy.ccc.blocks[zb].m_width;
        Pixy_mH[zb] = pixy.ccc.blocks[zb].m_height;
        Pixy_Sg[zb] = pixy.ccc.blocks[zb].m_signature;
      }
      // load frame rate to ensure camera has enough light
      PixyFps = pixy.getFPS();
    } else {BlockCnt = -1;} // flag an error
  }
//  t1[0] = micros(); // this function takes 680us whether we ask for 1 or 3 blocks
}

// --------------------------------------------------------------------------------

void Pixy_NumBlocks() {
  // If there are detect blocks, print them!
  if (pixy.ccc.numBlocks) {
    if (pixy.ccc.numBlocks != BlockCnt) {
//      Serial.println ("Detected " + String(pixy.ccc.numBlocks));
    }
//    for (int zI = 0;zI < pixy.ccc.numBlocks;zI++)
//    {
//      Serial.print("  block ");
//      Serial.print(zI);
//      Serial.print(": ");
//      pixy.ccc.blocks[zI].print();
//    }
  } else {
//    if (pixy.ccc.numBlocks != BlockCnt) {Serial.println("No blocks");}
  }
  BlockCnt = pixy.ccc.numBlocks;
}

// --------------------------------------------------------------------------------

void Pixy_ServoReset() {
  // if Pixy is connected then send command to set servos to default positions
  if (I2C_Pixy) {
    servoVal[0] = servoCtr[0]; servoVal[1] = servoCtr[1];
    pixy.setServos(servoVal[0],servoVal[1]);
  }
}

// --------------------------------------------------------------------------------
