// ################################################################################
//
//  PixyBot v0.0 (R1)
//
//  Released:  06/02/2022
//
//  Author: TechKnowTone
//
// ################################################################################
/*
    TERMS OF USE: This software is furnished "as is", without technical support, and
    with no warranty, expressed or implied, as to its usefulness for any purpose. In
    no event shall the author or copyright holder be liable for any claim, damages,
    or other liability, whether in an action of contract, tort or otherwise, arising
    from, out of or in connection with the software or the use or other dealings in
    the software.

    Micro:  ESP32

    This code controls a 4x4 robot chassis, with a Pixy2 camera mounted on it.
    Functions provided are as follows:
    * OLED Display mirroring to PC
    * Laser ranging & tracking
    * Pixy2 camera modes, including Pan & Tilt
    * WiFi remote control using Wii Nunchuk
    * Mecanum wheel drive demo
*/
// Declare Libraries
#include <Arduino.h>          // needed when using an ESP32 micro
#include <HardwareSerial.h>   // serial library
#include <Wire.h>             // I2C library
#include <FastLED.h>          // Neopixel library
#include "SSD1306Wire.h"      // OLED display library

#include <esp_now.h>        // ESP-NOW WiFi link library
#include <WiFi.h>           // ESP WiFi library
//#include <EEPROM.h>         // EEPROM flash access library
#include "SparkFun_VL53L1X.h" // VL53L1X laser ranger library
#include <Pixy2I2C.h>         // Pixy2 library for I2C connection

// Configuration
String Release = "PixyBot (Beta)";
String Date = "02/02/2022";

// I2C address map
// 0x29 VL53L1X laser range finder
// 0x3C OLED display
// 0x68 MPU 6050 3-axis motion sensor
// 0x54 Pixy2 I2C UART

// Display Settings
const int I2C_DISPLAY_ADDRESS = 0x3c;
const int SDA_PIN = 21;         // I2C SDA GPIO pin
const int SCL_PIN = 22;         // I2C SCL GPIO pin

// Initialize the oled display for address 0x3c
SSD1306Wire display(I2C_DISPLAY_ADDRESS, SDA_PIN, SCL_PIN);

// Create an instance of the Pix2 I2C UART interface
Pixy2I2C pixy;

// Define a task for core 0
TaskHandle_t Core0;

// Define constants
#define acc_1g  8192              // accelerometer 1g value
#define acc_calibration_valueX 0  // Enter the accelerometer calibration value, default 0
#define acc_calibration_valueY 0  // Enter the accelerometer calibration value, default 0
#define autoRestMax 32000 // autoRest max counter value
#define BatCal 383.9      // calibrated voltage multiplying factor
#define BatCritical 2534  // critical battery threshold @6.6v, default = 2672, set == 0 to ignore
#define BatMax 3148       // A0 for fully charged battery voltage, == 8.2v
#define BatPin 13         // analog input pin ADC15 on GPIO12
#define BatWarn 2687      // battery threshold of 7.0v gives low warning
#define DbLLX 119         // JoyX deadband lower limit
#define DbULX 136         // JoyX deadband Upper limit
#define DbLLY 119         // JoyY deadband lower limit
#define DbULY 136         // JoyY deadband Upper limit
#define gyro_address 0x68 // MPU-6050 I2C address (0x68 or 0x69)
#define Head_20 560       // fully left head angle, looking at left body corner
#define Head_90 1345      // mid point head angle, looking directly forward
#define Head_160 2130     // fully right head angle, looking at right body corner
#define INT_P34 34        // speed sensor interupt input FR
#define INT_P35 35        // speed sensor interupt input RL
#define INT_P36 36        // speed sensor interupt input FL
#define INT_P39 39        // speed sensor interupt input RR
#define INT_PIN 0         // VL53L1X GPIO interrupt to ADC input not actually used
#define LED_Pin 12        // assign GPIO12 for RGB LEDs
#define LtofDet 350       // max range for target detection in head scanning
#define LtofLimit 500     // Range max value is always limited to this figure
#define LtofMax 400       // out of range value for head scanning
#define LtofMin 55        // minimum range value
#define LtofStop 140      // range value at which autonomous movement stops
#define NumLEDs 13        // number of neopixel LEDs
#define PinBckLftA 33     // assign GPIO 26 to PWM rear left motor
#define PinBckLftB 25     // assign GPIO 27 to PWM rear left motor
#define PinBckRhtA  2     // assign GPIO 16 to PWM rear right motor
#define PinBckRhtB  4     // assign GPIO 17 to PWM rear right motor
#define PinFntLftA 26     // assign GPIO 33 to PWM front left motor
#define PinFntLftB 27     // assign GPIO 25 to PWM front left motor
#define PinFntRhtA 16     // assign GPIO  2 to PWM front right motor
#define PinFntRhtB 17     // assign GPIO  4 to PWM front right motor
#define PixWdMin 20       // minimum width for object recognition
#define PwmBits 8         // PWM counter width 8 bits, 0 - 255
#define PwmChBckLftA 0    // PWM channel assigned to rear left motor
#define PwmChBckLftB 1    // PWM channel assigned to rear left motor
#define PwmChBckRhtA 2    // PWM channel assigned to rear right motor
#define PwmChBckRhtB 3    // PWM channel assigned to rear right motor
#define PwmChFntLftA 4    // PWM channel assigned to front left motor
#define PwmChFntLftB 5    // PWM channel assigned to front left motor
#define PwmChFntRhtA 6    // PWM channel assigned to front right motor
#define PwmChFntRhtB 7    // PWM channel assigned to front right motor
#define PwmFreq 30000     // PWM frequency is 30 kHz
#define PWM_StartMax 45   // PWM to overcome start inertia
#define PWM_StartMin 30   // PWM to maintain movement
#define PxDbLL 152        // Pixy centre X deadband lower limit
#define PxDbUL 164        // Pixy centre X deadband upper limit
#define PyDbLL 100        // Pixy centre Y deadband lower limit
#define PyDbUL 108        // Pixy centre Y deadband upper limit
#define sw0Pin 14         // left switch SW0 assigned to GPIO14
#define sw1Pin 15         // right switch SW1 assigned to GPIO15
#define XSHUT_PIN 32      // GPIO32 controls VL53L1X XSHUT HIGH enable pin
 
// create a FastLED instance and define the total number of LEDs on the robot
CRGB LED[NumLEDs];  // LEDs connected to micro for o/p
CRGB LEDX[1];       // temp LED value
CRGB LM0[NumLEDs];  // LED memory bank 0
CRGB LM1[NumLEDs];  // LED memory bank 1

// create VL53L1X range sensor instance, and define its pins
// Note that INT_PIN is declared for the library, but not used in the code
SFEVL53L1X VL53L1X(Wire, XSHUT_PIN, INT_PIN);

// my Wii Transciever unique MAC: 50:02:91:68:F7:3F
// you need to build a Wii Transceiver and get its MAC address
uint8_t broadcastAddress[] = {0x50, 0x02, 0x91, 0x68, 0xF7, 0x3F};

// 250 byte max for Tx/Rx data
// It must match the receivers data structure
typedef struct ESPNOW_Buff250 {
    char ESPdata[250];
} ESPNOW_Buff250;

// Create an 8 byte message called NunchukDataTx, to hold I2C data messages
ESPNOW_Buff250 Tx_Buff;

// Create an 8 byte message to hold incoming sensor readings
ESPNOW_Buff250 Rx_Buff;

// define servo arrays
// [0] - Pixy2 Pan servo
// [1] - Pixy2 Tilt servo
//                     Pan,Tilt
int16_t servoCtr[] = { 554, 252}; // centre value sent to servos
int16_t servoInt[] = {   0,   0}; // servo tracking integrator
int16_t servoMax[] = { 947,1000}; // servo max pulses, Pan 60° left, Tilt leaning forward
int16_t servoMin[] = { 161,   0}; // servo min pulses, Pan 60° right, Tilt leaning backward
int16_t servoVal[] = { 554, 252}; // value sent to servos pan,tilt

// define Pixy arrays for 3 objects
// this avoids unnecessary error checking
int16_t Pixy_mX[3];
int16_t Pixy_mY[3];
int16_t Pixy_mW[3];
int16_t Pixy_mH[3];
int16_t Pixy_Sg[3];


// Declare and initialise global variables
int acc_cal_valueX = acc_calibration_valueX;   // Enter the accelerometer calibration value, default 1000
int acc_cal_valueY = acc_calibration_valueY;   // Enter the accelerometer calibration value, default 1000
int accelerometer_data_X;             // accelerometer data for the forward X axis
int accelerometer_data_Y;             // accelerometer data for the sideways Y axis
int accelerometer_data_rawX_i;        // instantaneous record for averaging
int accelerometer_data_rawX_OA;       // instantaneous record offset adjusted
int accelerometer_data_rawX_OALast;   // previous instantaneous record offset adjusted
int accelerometer_data_rawY_i;        // instantaneous record for averaging
int accelerometer_data_rawY_OA;       // instantaneous record offset adjusted
int accelerometer_data_rawY_OALast;   // instantaneous record offset adjusted
String Any$;                  // temporary string
long AnyLong;                 // any temp value
unsigned long anyMilli;       // any millisecond timer
int16_t anyVal;               // any temp variable
int16_t autoOFF;              // if set > 0 then counts down to servo switch OFF, if -1 disabled
int16_t autoRest;             // returns to a rest position after a timeout
int32_t BatAvg;               // preload averaging function with high value
int32_t BatDist;              // used battery discharge testing
int32_t BatSum;               // preload averaging sum with 20x max value
int32_t BatVol;               // instantaneous battery voltage
bool Blink;                   // LED blink clock
int16_t BlinkCnt;             // counter for LED Blink clock
int16_t BlockAsk;             // number of blocks asked for from the Pixy2 camera, default = 3
int16_t BlockCnt;             // latest number of blocks detected by the Pixy2 camera
int16_t BlockErr;             // records Pixy2 I2C request errors
bool Border;                  // if true then display has a border
uint8_t Bright = 32;          // FastLED brightness
uint8_t Brightness;           // Display brightness
int CalcCnt;                  // tracks the number of head calculations in autonomous mode
int16_t CamerLux;             // camera brightness, default 30
int16_t C_Cnt;                // counter used in 'C' button detection
byte C_Dn;                    // counter for 'C' button down period
byte C_Up;                    // counter for 'C' button up period
char cmdMode;                 // command mode
int16_t cmdSgn;               // cmdVal sign, 0 or -1
char cmdType;                 // command mode type
int cmdVal;                   // value associated with a cmdType
byte Col_B,Col_G,Col_R;       // current colours
int16_t CZ;                   // received button values
bool CZ_Wait;                 // flag used for button release wait function
int16_t DIR_FL;               // direction sign, +ve forward, -ve backwards
int16_t DIR_FR;               // direction sign, +ve forward, -ve backwards
int16_t DIR_RL;               // direction sign, +ve forward, -ve backwards
int16_t DIR_RR;               // direction sign, +ve forward, -ve backwards
bool DispBack;                // flag used to draw a back arrow on the display
bool DispChng;                // flag invokes initialisation code
bool DispClr;                 // set == true to clear the memory contents
int16_t DispCnt;              // counter used in display updating
int16_t DispCX,DispCY;        // display cursor coordinates
int16_t DispDel;              // >0 display delay counter, stops updates over I2C
bool DispDisp;                // set = true to immediately display memory contents
int16_t DispH;                // display rectangle width
byte DispMenu;                // group type, 0 = sensors, 1 = tasks, etc
int16_t DispMode;             // the type of display for a given mode
int16_t DispModeCnt;          // counter used in display updating
bool DispMon;                 // == true if connected to OLED Mirror app
bool DispNOP;                 // set == true to stop display I2C updates
int16_t DispNow;              // if > 0 display special modes
int16_t DispNxtCnt;           // counter used in auto-branching of display
int16_t DispPnt;              // points at the current display mode
int16_t DispTask;             // task pointer for display functions
int16_t DispTx;               // counter used with display mirror app
int16_t DispVal;              // general display value, passed from tasks
int16_t DispW;                // display rectangle width
int16_t Dither;               // overlay signal used to reduce stiction
bool DithON;                  // dither applied flag, true = on, false = off
int16_t Drive;                // if > 0 then driving in various directions
int16_t DriveLast;            // previous driving direction
int16_t DriveFL;              // front left-side PWM value
int16_t DriveFR;              // front right-side PWM value
int16_t DriveLft;             // left-side PWM value
int16_t DrivePWM;             // PWM drive value for all 4 wheels
int16_t DriveRht;             // right-side PWM value
int16_t DriveRL;              // rear left-side PWM value
int16_t DriveRR;              // rear right-side PWM value
int16_t DriveTmp;             // temp value used in drive modes
bool ESC;                     // true to escape all functions
int16_t Gear;                 // value used to determine PWM Max values
int gyro_Z_data_raw;          // raw Z gyro value read from MPU6050
bool HeadActive;              // = true if head is in scan mode for tasks
bool HeadAdj;                 // flag used in head scanning 
int HeadAng;                  // HeadVal converter to degrees
bool HeadAtt;                 // = true when head servo has been attached
int16_t HeadCnt;              // counter used in head movements
int16_t HeadCyc;              // cycle counter used in head movements
int16_t HeadDir;              // head movement direction, +1 == RH, -1 == LH
long HeadErr;                 // value used in head angle adjustment
bool HeadErrSt;               // trip flag used in head scanning
long HeadErrTrip;             // error tripping value used in head angle adjustment
long HeadErrTrk;              // error tracking value used in head angle adjustment
unsigned long HeadInterval;   // main loop head movement interval in microseconds
int16_t HeadMax;              // maximum servo value
unsigned long HeadMicros;     // main loop head movement timer in microseconds
int16_t HeadMin;              // minimum servo value
int16_t HeadMode;             // pointer used in head scanning
int16_t HeadPause;            // 20ms counter to temporarily pause head movement
int16_t HeadSubTask;          // head movement sub task pointer
int16_t HeadTask;             // head movement task pointer
int16_t HeadTaskLast;         // previous head movement task pointer
int16_t HeadTaskNext;         // head movement next task pointer
bool HeadTaskRun;             // set to true during head tasks
int16_t HeadTgt;              // target head servo angle
int16_t HeadTgtCnt;           // counter used in automatic head servo move to target actions
int HeadVal;                  // current head servo angle in microseconds
byte I2C_Error;               // flag returned from I2C comms
bool I2C_Disp;                // true if display is detected during POST
bool I2C_LTOF;                // true if VL53L1X is detected during POST
bool I2C_MPU;                 // true if MPU6050 is detected during POST
bool I2C_Pixy;                // true if Pixy2 is detected during POST
int16_t Inc = -1;             // increment value
bool Int_En;                  // interrupts enabled flag, == true if enabled
unsigned long IntLast[] = {0,0,0,0};  // previous interrupt times in micros()
volatile uint8_t IntMask;     // 4-bit mask used to flag the occurence of interrupts
volatile unsigned long IntMicros[4];   // interrupt times in micros()
bool  IntP34Ch;               // if == true then a slot transition has occured
int16_t IntP34Cnt;            // pulse count for pin INT_P34
int16_t Int34Rev;             // slot counter for pin INT_34, looking for 20 slots == 1 revolution
int16_t Int34RPM;             // calculated rpm for INT_34 slot, actually rpm x 10
bool IntP34State;             // latest state of the INT_P34 input HIGH/LOW
bool  IntP35Ch;               // if == true then a slot transition has occured
int16_t IntP35Cnt;            // pulse count for pin INT_P35
int16_t Int35Rev;             // slot counter for pin INT_35, looking for 20 slots == 1 revolution
int16_t Int35RPM;             // calculated rpm for INT_35 slot, actually rpm x 10
bool IntP35State;             // latest state of the INT_P35 input HIGH/LOW
bool  IntP36Ch;               // if == true then a slot transition has occured
int16_t IntP36Cnt;            // pulse count for pin INT_P36
int16_t Int36Rev;             // slot counter for pin INT_36, looking for 20 slots == 1 revolution
int16_t Int36RPM;             // calculated rpm for INT_36 slot, actually rpm x 10
bool IntP36State;             // latest state of the INT_P36 input HIGH/LOW
bool  IntP39Ch;               // if == true then a slot transition has occured
int16_t IntP39Cnt;            // pulse count for pin INT_P39
int16_t Int39Rev;             // slot counter for pin INT_39, looking for 20 slots == 1 revolution
int16_t Int39RPM;             // calculated rpm for INT_39 slot, actually rpm x 10
bool IntP39State;             // latest state of the INT_P39 input HIGH/LOW
unsigned long IntLast40[4];   // revolution period in ms, 20 slots, 40 counts
unsigned long IntPeriods[4];  // inter-slot periods in µs
bool JoyActive;               // flag used to prevent stack overflow during moves
int16_t JoyD;                 // direction value 0 - 8
int16_t JoyDL;                // direction 1 counter
int16_t JoyMode;              // direction of travel flag, 0 = stationary, 1 = forward, -1 = backward
int16_t JoySpd;               // speed vector = max(JoyX,JoyY)
int16_t JoyV;                 // joystick vector  = SQRT(SQ(JX) + SQ(JY))
int16_t JoyX;                 // received value
int JoyXV;                    // converted value 0 - 127
int16_t JoyXC;                // received value offset
int16_t JoyY;                 // received value
int16_t JoyYC;                // received value offset
int JoyYi;                    // tracking filter for received JoyY value
int JoyYV;                    // converted value 0 - 128
boolean JoyYT;                // tracking filter flag; true if not on track
int JX;                       // global variable used with Joystick app
int JY;                       // global variable used with Joystick app
char keyChar;                 // any keyboard character
int16_t keyVal;               // any keyboard value
int16_t LED_Cnt;              // counter used in LED sequencing
int16_t LED_Del;              // delay timer used in LED functions
int16_t LED_Max;              // max value used in LED functions
int16_t LedMode;              // a value > 0 determines LED flash sequence; if = 0 disabled
unsigned long LED_Period;     // LED timer period, default = 20ms
bool LED_Run;                 // == true if LED_Task is to run
bool LEDshow;                 // LED show flag
int16_t LedSpd;               // motor speed value used by LED animation
int16_t LED_Task;             // LED task pointer
bool LedTick;                 // set true during LED cycles
unsigned long LED_Timer;      // 20+ms LED timer
byte LEDVal;                  // values used in LED tasks
bool Lights;                  // if == true then turn Pixy2 lights ON
unsigned long loopDelay;      // RTOS task timer
bool LTOFcalc;                // if == true then calculate speed of ranged object in m/s
int16_t LTOFfps;              // ranger sensor measurement rate in fps
int16_t LTOFperiod;           // range sensor inter-measurementg period in millis()
int16_t LTOFspd10;            // speed of ranged target in m/s x 10
int16_t LTOFspd10L;           // previous speed of ranged target in m/s x 10
int16_t LTOFspdCnt;           // speed peak detector sustain limit
int16_t LTOFt0,LTOFt1;        // range sensor timers for fps
int16_t MainMode;             // PixyBot main state, 0 = sleep, 1 = Drive, 2 = LTOF, 3 = Pixy2
bool MainRun;                 // set true when a MainTask is running to prevent re-entry
int16_t MainTask;             // main task pointer
int16_t MainTaskNext;         // the task to run after an arrest; default is 0
int16_t ModeSel;              // MainMode selection
int16_t MoveState;            // move state machine pointer
String myMAC;                 // MAC address if this device
unsigned long next16ms;       // 16.666ms loop timer
unsigned long next40ms;       // 40ms loop timer
unsigned long period16ms;     // a period which varies between 16-17ms to acjhieve 60Hz
uint8_t phase16ms;            // a 3-phase pointer which adjusts the 16ms value to get 16.6666
int16_t Ping;                 // Rx 'ping' counter
bool PixyBotActive;           // = true if active, otherwise false
uint8_t PixyDel;              // delay timer used in checking for Pixy2
bool PixyEn;                  // == true when I2C to Pixy2 is active
uint8_t PixyFps;              // camera frames per second, indicates light level quality
int16_t PixyW;                // object width variable
unsigned long print40ms;      // 40ms variable print timer
int16_t PrintTgt;             // 0 == serial port, 1 == WiFi ESP-NOW
String PrintTx = "";          // printed strings
uint8_t PwmBckLftA;           // PWM value assigned to rear left A pin
uint8_t PwmBckLftB;           // PWM value assigned to rear left B pin
int16_t PwmBckMax;            // max reverse PWM used in LTOF demo
bool PwmEn;                   // if true motor PWM is enabled
int16_t PWM_FL;               // PWM applied to motor drive +/-255
int16_t PWM_FLTgt;            // target motor PWM drive +/-255
int16_t PWM_FR;               // PWM applied to motor drive +/-255
int16_t PWM_FRTgt;            // target motor PWM drive +/-255
int16_t PwmFwdMax;            // max forward PWM used in LTOF demo
int16_t PwmInt;               // integrator used in motor control
int16_t PWM_RL;               // PWM applied to motor drive +/-255
int16_t PWM_RLTgt;            // target motor PWM drive +/-255
int16_t PWM_RR;               // PWM applied to motor drive +/-255
int16_t PWM_RRTgt;            // target motor PWM drive +/-255
int16_t PWM_Start;            // value needed to ensure motor start
int16_t R0,R1,R2,R3,R4,R5,R6,RM; // range values for backaway and target tracking
int Range;                    // range value limited to RangeMax
int RangeRaw;                 // range value obtained from VL53L1X
int16_t RangeAng;             // angle at which minimum range was measured
int16_t RangeCentre;          // distance value looking forward
int RangeDiff;                // difference in RangeRaw from RangeLast
bool RangeEn;                 // == true if VL53L1X ranging is active
bool RangePnt;                // true when a centre range value is recorded, or extremes of sweep
int16_t RangeCentreTracker;   // slugged distance value looking forward
long RangeIntL;               // L-H integrator used in head centring
long RangeIntR;               // R-H integrator used in head centring
int RangeLast;                // previous range measurement
int16_t RangeLeft;            // distance value lookin left
bool RangeLL;                 // == true if range < RangeMin
int RangeMax;                 // the current limited to Range measurements
int RangeMin;                 // the current lower limit of Range measurements
unsigned long RangeMinVal;    // min range value during close range scanning 
int16_t RangeMinAng;          // angle at which RangeMin occured
int16_t RangeMinMem;          // RangeMin at end of each half sweep
int16_t RangeRight;           // distance value lookin right
int16_t RangeTrig;            // range trigger point used in head scanning
bool RangeUL;                 // == true if range exceeds RangeMax
bool readWiiCall;             // flag prevents call stack overflow
int16_t Rx_Chk;               // Rx checksum
int16_t RxCZ;                 // buffered received CZ value
int16_t RxJoyX;               // buffered received JoyX value
int16_t RxJoyY;               // buffered received JoyY value
int16_t Rx_len;               // length of WiFi received data block
int16_t Rx_Pnt;               // data pointer user by Rx state machine
bool RxRec;                   // set true when a valid frame of data is received
int16_t RxState;              // receiver state machine state
int16_t Rx_Task;              // task pointer user by Rx state machine
int16_t RTxTimeout;           // counter used to auto-reset if WiFi fails
int16_t RxVal;                // value received from serial Rx
byte RxWiFi[6];               // array holding 6 bytes of Wii received data
bool SerialRx;                // == true if any character has been received
int16_t ServoInc;             // servo test increment direction
int16_t ServoPnt;             // servo target pointer for manual control
uint16_t ServoVal;            // value sent to both Pixy2 servos over I2C, default = 500
int16_t SetSel;               // setting selection pointer
int16_t SigAp;                // block array pointer to SigId data
int16_t SigHi;                // object signtaure ID width trigger point, large
int16_t SigId;                // object of interest signtaure ID, default is 0 (largest)
int16_t SigLo;                // object signtaure ID width trigger point, small
int16_t SigTask;              // object of interest recognition task pointer
bool SlotEn;                  // if == true then constantly monitor wheel slot encoders
uint8_t Snore;                // snore animation character
String Status$ = "";          // any status message to be displayed on OLED
int16_t Steer;                // value applied when steering left/right
int16_t SubCnt;               // sub task general counter
int16_t SubDel;               // subtask delay counter
int16_t SubNext;              // pointer to next sub task
int16_t SubRpt;               // subroutine repeat counter
int16_t SubTask;              // sub task pointer
int16_t SubTask1;             // level 1 sub task pointer, used when Pixy2 Pan/Tilt is using SubTask
int16_t SubVal;               // anyvalue used in a subroutine
int16_t SubWait;              // SubTask wait period
int16_t sw0Cnt;               // button switch counter
int16_t sw0DwnTime;           // button pressed down time
bool sw0LastState;            // previous state of button switch, HIGH/LOW
bool sw0_Nop;                 // if == true don't perform switch functions
bool sw0State;                // state of read button switch pin
int16_t sw0Timer;             // timer used to detemine button sequences
bool sw0Wup;                  // set == true for SW0 to wait for button release
int16_t sw1Cnt;               // button switch counter
int16_t sw1DwnTime;           // button pressed down time
bool sw1LastState;            // previous state of button switch, HIGH/LOW
bool sw1_Nop;                 // if == true don't perform switch functions
bool sw1State;                // state of read button switch pin
int16_t sw1Timer;             // timer used to detemine button sequences
bool sw1Wup;                  // set == true for SW1 to wait for button release
byte t;                       // any timer pointer
int t0[4];                    // any function timer
int t1[4];                    // any function timer
int t2[4];                    // any function timer
int16_t Task16ms;             // 16.666ms subtask pointer
int16_t Task40ms;             // 40ms subtask pointer
bool TEST = false;            // set == true during assembly, then false for normal use
int16_t TEST_PWM;             // +/- PWM value used during TEST mode
int16_t TgtCnt;               // counter used in automatic target and drive actions
int16_t TgtMovCnt;            // counter used in automatic target and drive actions
int16_t Tx_PingCnt;           // 1 second Tx ping timer
bool USB;                     // == true if initial battery is below 5.5v
uint8_t VL53L1X_OC;           // VL53L1X ROI SPAD centre
uint16_t VL53L1X_ROIX;        // VL53L1X ROI SPAD width
uint16_t VL53L1X_ROIY;        // VL53L1X ROI SPAD height
unsigned long VL53L1X_Skip;   // read function skip timer
unsigned long VL53L1X_SkipDel;// read function skip timer delay tracker
int16_t VL53L1X_Task;         // if > 0 then run VL53L1X tasks
unsigned long VL53L1X_Timer;  // general timer
int16_t WiFiCntC;             // C button counter for WiFi enable
int16_t WiFiConCnt;           // counter used in connection retry times
bool WiFiConnected;           // == true if a WiFi connection has been established
bool WiFiEn;                  // true when WiFi is enabled with prolonged 'Z' button activity
int16_t WiFiPing;             // Rx counter used to maintain connected state
int16_t WiFiRestCnt;          // counter used to goi to rest if no WiFi demands
bool WiFiRx;                  // true when a block of data has been received
long WiFiRxBytes;             // number of bytes received over the WiFi link
int16_t WiFiRxErr;            // number of bytes received over the WiFi link
bool WiFiRxRec;               // true when a block of data has been successfully received
int16_t WiFiTryCnt;           // WiFi try to connect count on error count
int16_t WiFiTryNum;           // WiFi try to connect total count
bool WiFiTx;                  // true when a block of data has been sent successfully
int16_t WiFiTx_CB;            // == 1 call back is clear to send, == -1 resend
int16_t WiFiTx_Chk;           // Tx checksum
int16_t WiFiTx_len;           // >0 when WiFi Tx buffer contains data to be sent
long WiFiTxBytes;             // number of bytes received over the WiFi link
int16_t WiFiTxErr;            // WiFi data error
int16_t WiFiTxErrCnt;         // WiFi Tx error count, leads to disconnection
int16_t WiiType;              // device type, 0 == Nunchuk, 1 == Classic
byte Z_Dn;                    // button pressed state; 0 == UP, >= 1 == Down
bool Z_Mode;                  // joystick motion flag, normally false, if == true then slide mode

// ---------------------------------------------------------------------------------

void setup() {
  // Setup code, runs once:

  // Initialise PWM pins and output states
  pinMode(PinBckLftA, OUTPUT); digitalWrite(PinBckLftA,LOW);
  pinMode(PinBckLftB, OUTPUT); digitalWrite(PinBckLftB,LOW);
  pinMode(PinBckRhtA, OUTPUT); digitalWrite(PinBckRhtA,LOW);
  pinMode(PinBckRhtB, OUTPUT); digitalWrite(PinBckRhtB,LOW);
  pinMode(PinFntLftA, OUTPUT); digitalWrite(PinFntLftA,LOW);
  pinMode(PinFntLftB, OUTPUT); digitalWrite(PinFntLftB,LOW);
  pinMode(PinFntRhtA, OUTPUT); digitalWrite(PinFntRhtA,LOW);
  pinMode(PinFntRhtB, OUTPUT); digitalWrite(PinFntRhtB,LOW);

  // initialise other pins
  pinMode(BatPin,INPUT);        // ADC2_4  GPIO13 battery divider
  pinMode(INT_P34,INPUT);       // speed sensor interupt pin front-right
  pinMode(INT_P35,INPUT);       // speed sensor interupt pin rear-left
  pinMode(INT_P36,INPUT);       // speed sensor interupt pin front-left
  pinMode(INT_P39,INPUT);       // speed sensor interupt pin rear-right
  pinMode(sw0Pin,INPUT_PULLUP); // button switch sw0 Pin
  pinMode(sw1Pin,INPUT_PULLUP); // button switch sw1 Pin
  pinMode(XSHUT_PIN,OUTPUT); digitalWrite(XSHUT_PIN,LOW);     // switch VL53L1X OFF
  
  setDefaults();  // reset variables

  Serial.begin(115200);

  Wire.begin();

  // initialise the FastLED component
  FastLED.addLeds<WS2812B, LED_Pin, GRB>(LED, NumLEDs);
  FastLED.setBrightness(Bright);

  // initialise the eight PWM channels, setting each to max
  for (int zP = 0;zP < 8;zP++) {
    PwmSetup(zP);
    PwmAttach(zP);
    ledcWrite(zP, 0);
  }

  // grab initial battery voltage before WiFi kills the ADC2 function
  // take 10 readings over a 10ms period
  for (int zI = 0;zI < 10;zI++) {BatVol+= analogRead(BatPin); delay(1);}
  BatVol/= 10;
    
  // initialise WiFi link
  WiFi.mode(WIFI_STA);
  myMAC = WiFi.macAddress();
  Serial.println("my MAC= " + String(myMAC));
  WiFi.disconnect();
  
  // Set ESP-NOW Role as bidirectional
//  esp_now_set_self_role(ESP_NOW_ROLE_COMBO);

  // Init ESP-NOW and returns its status
  if (esp_now_init() != ESP_OK) {
    Serial.println("Error initializing ESP-NOW");
    return;
  }

  // Once ESPNow is successfully Init, we will register for Send CB to
  // get the status of Trasnmitted packet
  esp_now_register_send_cb(OnDataSent);

  // Register peer
  esp_now_peer_info_t peerInfo; //initialize and assign the peer information as a pointer to an addres
  memcpy(peerInfo.peer_addr, broadcastAddress, 6); //copy the value of  broadcastAddress with 6 bytes to peerInfo.peer_addr
  peerInfo.channel = 0;  //channel at which the esp talk. 0 means undefined and data will be sent on the current channel. 1-14 are valid channels which is the same with the local device 
  peerInfo.encrypt = false; //not encrypted

  //Add the device to the paired device list 
  if (esp_now_add_peer(&peerInfo) != ESP_OK){ 
    Serial.println("Failed to add peer");
    return;
  }

  // Register for a callback function that will be called when data is received
  esp_now_register_recv_cb(OnDataRecv);

  RunPOST();  // initialise I2C devices, set intro display, etc

  synchLoopTimers(0);
  
  // define the Core 0 task parameters
  // Core 0 is normally used for WiFi, but here we use it to read sensors
  // leaving Core 1 to create the LED patterns and clock them out
  // Note, this code starts and runs immediately, before loop() does
  xTaskCreatePinnedToCore(
    loop0,       // Task function.
    "Core0",     // name of task.
    10000,       // Stack size of task
    NULL,        // parameter of the task
    1,           // priority of the task
    &Core0,      // Task handle to keep track of created task
    0);          // pin task to core 0
}

// ---------------------------------------------------------------------------------

  // the controller will react to a soft RESET event
  // this function effectively restarts the code as if from power-up
void(* resetFunc) (void) = 0; //declare reset function at address 0

// ---------------------------------------------------------------------------------

void loop() {
  // Main loop code, runs repeatedly.
//  t1[0] = micros(); // end time
//  t0[1] = micros(); // start time

  //###############################################################################
  //
  //  WiFi comms check & respond
  //
  //###############################################################################
  // if WiFi is conencted then check Rx buffer on every cycle, so that we respond to
  // received packets as soon as possible
  if (WiFiConnected) {
    if (WiFiRx) {readWiFiRx();} // WiFi Rx buffer contains unread data so deal with it before adding new
  }

  readSerial();                 // check serial port on every loop cycle

  if (VL53L1X_Task > 0) {VL53L1X_Run();}  // perform laser measurements


  //###############################################################################
  //
  //  RGB LED 20+ms Tasks (50Hz variable)
  //
  //###############################################################################
  // The LED tasks are run at a variable rate
  if (LED_Run) {LED_Main();}
  if ((millis() - LED_Timer) >= LED_Period) {
    LED_Timer = millis();
    if (LEDshow) {FastLED.show(); LEDshow = false;} // update LEDs if LEDshow == true
    LED_Run = true; // initiate the next LED_Task process
    LedTick = true; // trigger timed tasks
  }

  
  //###############################################################################
  //
  //  16.666ms Tasks (60Hz)
  //
  //###############################################################################
  // These tasks are run in a consequitive ripple to reduce the overhead on more
  // critical code. They are performed in reverse case order and will all complete
  // within the 16.666ms period
  if (Task16ms > 0) {
    if (TEST) {
      switch(Task16ms) {
        case  1: PWM_Test(); break;   // perform simple PWM task
        case  2: readSW1(); break;    // read right button switch SW1
        case  3: readSW0(); break;    // read left button switch SW0
      }
    } else {
      switch(Task16ms) {
        case  1: MotorDriveTask();                      // control motor PWM and dither signals
        case  2: readSW1(); break;                      // read right button switch SW1
        case  3: readSW0(); break;                      // read left button switch SW0
        case  4: if (WiFiConnected) {readWii();} break; // if connected to WiFi read Wii data and respond
        case  5: // if WiFi fails stop movement and trigger autoRest
          // if readWiiCall == true then we are currently locked in a movement, so ignore timeout
          if (PixyBotActive && (RTxTimeout > 0) & !readWiiCall) {
            RTxTimeout--;
            if (RTxTimeout == 1) {
              Serial.println("WiFi failed!...");
              LedMode = 0; autoRest = 0;
            }
          } break;
        case 6:
          //###############################################################################
          //
          //  Main tasks (16.666ms 60Hz)
          //
          //###############################################################################
//          Serial.println(t1[0] - t0[0]);
          switch (MainTask) {
          case  0: break; // default null task
          case 10: // Drive mode default, do nothing here
            break;
          case 11: // Drive mode Demo 1
            MainTask_Drive_Demo1(); break;
          case 21: // goto PixyBotActive static position, LTOF active
            MainTask_LTOF(); break;
          case 22: // backaway from target mode, then return to start
            MainTask_TrackWall(0); break;
          case 23: // track wall target mode
            MainTask_TrackWall(1); break;
    //      case 4: // autonomous demo mode - to be written
    //        MainTask_Autonomous(); break;
          case 31: // Pixy2 static camera demo
            MainTask_PixyStatic(); break;
          case 32: // Pixy2 pan & tilt demo
            MainTask_PixyPanTilt(0); break;
          case 33: // Pixy2 object follower turning
            MainTask_PixyFollow(0); break;
          case 34: // Pixy2 object follower moderate drive
            MainTask_PixyFollow(1); break;
          case 35: // Pixy2 object follower full drive
            MainTask_PixyFollow(2); break;
          case 36: // Pixy2 object follower full drive
            MainTask_PixyGoalie(); break;
          case 40: // Settings selection, do nothing here
            break;
          case 41: // Camera settings selection
            MainTask_Set_Camera(); break;
          case 99: // Motor auto-controller for slot counting
            MainTask_MotorAutoCnt(); break;
          } break;
        case 7: if (PixyEn) {LoadPixyBlocks();} break;  // fill an array with Pixy2 blocks
        case 8: if (PixyEn) {BlockErr = Pixy_GetBlocks();} break;   // ask Pixy2 for blocks
       }
    }
    Task16ms--;
  } else if ((millis() - next16ms) >= period16ms) {
      // reset 16.666ms timer and task pointer
      next16ms = millis(); Task16ms= 8;
      // to get 16.666ms we use a sequence of 16,17,17,16,17,17...
      phase16ms++; if (phase16ms > 2) {period16ms = 16; phase16ms= 0;} else {period16ms = 17;}
  }

  
  //###############################################################################
  //
  //  40ms Tasks (25Hz)
  //
  //###############################################################################
  // These tasks are run in a consequitive ripple to reduce the overhead on more
  // critical code. They are performed in reverse case order and will all complete
  // within the 40ms period
  if (Task40ms > 0) {
    if (TEST) {
      // tasks performed during TEST mode
      switch(Task40ms) {
        case 1: if (I2C_Disp) {Display_40ms();} break;    // update display if necessary
        case 2: if (I2C_Pixy) {Pixy_Servo_Test();} break; // simple Pixy2 servo sweep
        case 3: if (I2C_Pixy) {Pixy_NumBlocks();} break;  // ask Pixy2 for blocks
        case 4: if (I2C_Pixy) {Pixy_GetBlocks();} break;  // ask Pixy2 for blocks
        case 5: if (!I2C_Pixy) {Pixy_Check();} break;     // keep checking for Pixy2
        case 6: RGB_Test(); break;                        // cycle LED colours
        case 7: if (I2C_MPU) {MPU_read_Rotation();} break;// read Z gyro
        case 8: if (I2C_MPU) {MPU_read_AccelAll();} break;// read accelerometer values
      }
    } else {
      // normal PixyBot2 tasks
      switch(Task40ms) {
        case 1: if (I2C_Disp) {Display_40ms();} break;              // update display if necessary
        case 2: if (!I2C_Pixy) {Pixy_Check();} break;               // keep checking for Pixy2
        case 3: if (Ping > 0) {Ping--;} break;                      // time-out data received flag
        case 4: // if not connected to WiFi run then try to connect every second
          if (!WiFiConnected) {
            // only try to connect WiFi if no 'Pings' are being received
            if (Ping == 0) {
              WiFiConCnt--; if (WiFiConCnt < 1) {
                WiFiConCnt = 50; WiFiTryToConnect();
  //              Serial.println("Trying to connect WiFi");
              }
            }
          } break;
        case 5:
          if (WiFiPing > 0) {
            WiFiPing--;
            if (WiFiPing < 1) {
               // we have stopped receiving so disconnect WiFi
              WiFiDisconnect();
  //            Serial.println("WiFiPing == 0");
            }
          } break;
      }
    } Task40ms--;
    // check for slack time
//    if (Task40ms == 0) {Serial.println("Free " + String(40 - (millis() - next40ms)));}
  } else if ((millis() - next40ms) >= 40) {
      // reset timer and task pointer to the first task
      next40ms = millis(); Task40ms = 5; if (TEST) {Task40ms = 8;}
      // check display mirror auto time-out
      if (DispTx > 0) {DispTx--; if (DispTx == 0) {DispMon = false;}}
  }


  //###############################################################################
  //
  //  40 ms Print Task - Variable
  //
  //###############################################################################
  // this code uses either the serial port or the WiFi link to send messages
  // at 115200 baud we can send 11.5 chars/ms, so 64 chars takes 5.5ms.
  if (millis() >= print40ms) {
    print40ms = millis() + 40;
    if (PrintTx.length() > 0) {
      // characters in string buffer so send some/all of them
      print40ms = millis() + 6; // shorten the delay for more data to send
      Tx_PingCnt = 0;           // reset the WiFi ping counter
      if (PrintTgt == 0) {
        // default uses the serial port
        if (Serial.availableForWrite() >= 64) {
          if (PrintTx.length() <= 64) {
            Serial.print(PrintTx);
            PrintTx = "";   // empty the buffer
          } else {
            Serial.print(PrintTx.substring(0,64));
            PrintTx = PrintTx.substring(64);
          }
        }
      } else {
        // text data has been received from WiFi so respond using ESP_NOW
        // ensure that we have had a call back from the previous frame
        if (WiFiConnected) {
          // only trannsmit PrintTx data if connected
          if (WiFiTx_CB == 1) {
            WiFiTx_CB = 0;  // clear the call back flag when sending
            if (PrintTx.length() <= 32) {
              WiFiTx_len = PrintTx.length();
              PrintTx.toCharArray(Tx_Buff.ESPdata,WiFiTx_len + 1);
              WiFiTxBytes += WiFiTx_len;  // track number of bytes sent
              WiFiTxGetChecksum(WiFiTx_len); Tx_Buff.ESPdata[WiFiTx_len] = WiFiTx_Chk; WiFiTx_len++;
//              t0 = micros();
              esp_now_send(broadcastAddress, (uint8_t *) &Tx_Buff, WiFiTx_len);
              PrintTx = "";   // empty the buffer
            } else {
              WiFiTx_len = 32;
              Any$ = PrintTx.substring(0,WiFiTx_len);
              Any$.toCharArray(Tx_Buff.ESPdata,WiFiTx_len + 1);
              WiFiTxBytes += WiFiTx_len;  // track number of bytes sent
              WiFiTxGetChecksum(WiFiTx_len); Tx_Buff.ESPdata[WiFiTx_len] = WiFiTx_Chk; WiFiTx_len++;
//              t0 = micros();
              esp_now_send(broadcastAddress, (uint8_t *) &Tx_Buff, WiFiTx_len);
              PrintTx = PrintTx.substring(32);
            }
          } else if (WiFiTx_CB == -1) {
            // an error has occured so resend the data
            if (WiFiTryCnt < 5) {
              WiFiTryCnt++;   // track the number of retries
              WiFiTx_CB = 0;  // clear the call back flag when sending
//              t0 = micros();
              esp_now_send(broadcastAddress, (uint8_t *) &Tx_Buff, WiFiTx_len);
            } else {
              // too many resends so disconnect WiFi link
              WiFiDisconnect();
              Serial.println("Failed to Tx!");
            }
          }
        } else {
          // not connected so clear PrintTxdata
          PrintTx = "";
        }
      }
    }
  }
//  timing test functions to see how much slack space is available
//  t1[1] = micros(); // end time
//  t = 0; if ((t1[t] - t0[t]) > t2[t]) {t2[t] = t1[t] - t0[t]; Serial.println("0 Core 1 System " + String(t2[t]));}
//  t = 1; if ((t1[t] - t0[t]) > t2[t]) {t2[t] = t1[t] - t0[t]; Serial.println("1 Loop() " + String(t2[t]));}
//  t = 2; if ((t1[t] - t0[t]) > t2[t]) {t2[t] = t1[t] - t0[t]; Serial.println("2 Core 0 System " + String(t2[t]));}
//  t = 3; if ((t1[t] - t0[t]) > t2[t]) {t2[t] = t1[t] - t0[t]; Serial.println("3 For(;;) " + String(t2[t]));}
//
//  t0[0] = micros(); // start time
}

// ---------------------------------------------------------------------------------

void loop0(void * pvParameters) {
  // this function runs on Core 0, also used for ESP32 WiFi NOW
  // *** WARNING *** do not use for time consuming code like display, I2C, FastLED, etc
  // the following message, sent once at boot, confirms core 0 assignment
  Serial.print("Core0 task running on core "); Serial.println(xPortGetCoreID());

  for(;;){
    // infinite loop, just like loop() but running in Core 0
    // you must never use 'break' or 'return' statement here or it will crash
    // we tell the free RTOS system to call this function every 1 ms
  
//    t1[2] = micros(); // end time
//    t0[3] = micros(); // start time

  //###############################################################################
  //
  //  Read slot sensors
  //
  //###############################################################################
    if (SlotEn) {Read_Slots();}

//    t1[3] = micros(); // end time
//    t0[2] = micros(); // start time

    // return to RTOS system with a 1ms call back, should allow rpm sensing <750rpm
    vTaskDelay(1/portTICK_PERIOD_MS);
  }
}

// --------------------------------------------------------------------------------

void OnDataRecv(const uint8_t * mac, const uint8_t *incomingData, int len) {
  // Callback when data is received over Wi-Fi
  // we copy the data into a buffer and set a flag so that the main loop will act
  // to handle it
  while (WiFiRx) {
    // buffer contains unread data so deal with it before adding new
    readWiFiRx();
  }

  memcpy(&Rx_Buff, incomingData, len);

  // incoming data will always be two or more characters, ie. text == $+++
  if (len < 2) {Serial.println("Rx Error - no Rx data"); return;}
  
  if (Rx_len > 0) {Serial.println("Rx Error - buffer in use");} // this should not happen

  // if connected Rx data will have a checksum so check it first
  Rx_len = len - 1; Rx_Pnt = 0;
  RxGetChecksum(Rx_len);  // generate a checksum from Rx data
  if (Rx_Chk != Rx_Buff.ESPdata[len - 1]) {
    Serial.println("Rx Error - checksum");
  }
  // set flags before exiting
  WiFiRxBytes += Rx_len; Rx_Task = 0; WiFiRx = true;
  WiFiPing = 50;    // set link time-out to 2 seconds, Wii Transceiver will keep setting this with Wii data
  WiFiRxRec = true; // set flag indicating data has been recieved from Wii Transceiver
//    Serial.print("Bytes received: "); Serial.println(len);
}

// --------------------------------------------------------------------------------

void OnDataSent(const uint8_t *mac_addr, esp_now_send_status_t status) {
  // Callback when data is sent over Wi-Fi, determines success or failure
//  t1 = micros();
//  Serial.print("Send Status: ");
  if (WiFiConnected) {
    // the link is already active so handle normal success/failure scenarios
    if (status == 0){
//      Serial.println("Success " + String(t1-t0));
      WiFiTx = true;    // flag WiFi Tx success
      WiFiTx_CB = 1;    // set flag to indicate that a call back state has occured
      WiFiTxErrCnt = 0; // clear the error flag in case of a glitch
      WiFiPing = 50;    // set link time-out to 2 seconds
    } else {
//      Serial.println("Fail " + String(t1 - t0));
      WiFiTx = false;   // flag WiFi Tx failure
      WiFiTx_CB = -1;   // set flag to indicate that a resend is required
      WiFiTxErrCnt++;
    }
  } else {
    // not connected so we must have tried to connect and may get an error
    if (status == 0){
      // successful so enter the connected state
      Serial.println("Connected!");
      WiFiConnected = true;
      WiFiPing = 50;    // set link time-out to 2 seconds, Wii Transceiver will keep setting this with Wii data
      WiFiTxErrCnt = 0; // reset error count for this connection
    } else {
//      Serial.println("Status = " + String(status));
      WiFiTxErrCnt++;
    }
    WiFiTx_CB = 1;
  }
}

// ---------------------------------------------------------------------------------

void RunPOST() {
  // start-up initialisation processes, mainly testing I2C addresses
  // 0x3C OLED display
  // 0x68 MPU 6050 3-axis motion sensor
  // 0x29 VL53L1X laser range finder
  // 0x54 Pixy2 I2C UART
  Serial.println("\n\n" + Release + "  " + Date + "\n");
  
  // Print TEST mode?... holding down SW1 during boot forces this condition
  if (!TEST) {if (digitalRead(sw1Pin) == LOW){TEST = true;}}
  if (TEST) {Serial.println("TEST mode active");}

  // we would normally set up battery quick averaging if we could continue to read ADC2
  // so in this version I report things on the serial port at bootup and the OLED
  // display also indicates initial readings. So you just have to press RESET to get a value
  BatAvg = BatVol; BatSum = BatVol * 50;  // preload the battery sum for averaging
  AnyLong = (BatAvg - BatCritical)*100/(BatMax - BatCritical);
  if (AnyLong > 100) {AnyLong = 100;} // limit max to 100%
  if (AnyLong < 0) {AnyLong = 0;}     // limit min to 0%
  Serial.println("Battery " + String(BatVol) + "\t" + String(float(BatAvg)/BatCal) + "v\t" + String(AnyLong) + "%");
  randomSeed(BatVol); // use the battery voltage as the random seed factor
  // if battery is less than 5.5v then USB mode is assumed, which makes servos inoperable
  if (BatAvg <= 2111) {USB = true;}
  
  // Test for display and initialise
  Wire.beginTransmission(0x3C);
  I2C_Error = Wire.endTransmission();
  if (I2C_Error == 0) {
    I2C_Disp = true; Display_Init();
    Display_Intro();
    Serial.println("Display initialised");
  } else {I2C_Disp = false; Serial.println("Display Not Found!");}

  // initilaise RGB LED array
  FastLED.clear();
  FastLED.show(); // This sends the updated pixel color to the hardware.
  Serial.println("RGB LEDs cleared.");

  // Test for MPU 6050 3-axis motion sensor and initialise
  Wire.beginTransmission(0x68);
  I2C_Error = Wire.endTransmission();
  if (I2C_Error == 0) {
    Serial.println("Initialising Gyros...");
     I2C_MPU = true; MPU_6050_Initialise();
  } else {I2C_MPU = false; Serial.println("MPU6050 Not Found!");}

  // Test for VL53L1X laser sensor and initialise
  digitalWrite(XSHUT_PIN,HIGH);  // enable VL53L1X sensor
  Wire.beginTransmission(0x29);
  I2C_Error = Wire.endTransmission();
  if (I2C_Error == 0) {
    I2C_LTOF = true; Serial.println("Initialising VL53L1X...");
    VL53L1X_ON();   // turn it ON briefly to confirm that it is there
    if (!TEST) {VL53L1X_OFF();}  // turn the laser OFF by default if not in TEST mode
  } else {I2C_LTOF = false; Serial.println("VL53L1X Not Found!");}

  // Test for Pixy2 camera on I2C, we do this last to give Pixy2 time to boot, but
  // it is unlikely that it will be seen by POST, so we also do it as a task
  Pixy_Check();

  if (TEST) {
    Bright = 8; FastLED.setBrightness(Bright);  // don't have the LEDs too bright
  } else {
    // normal startup
    SetMainMode(0);
    display.setBrightness(Brightness);  // have a bright display for the intro
    DispDel = 50; // wait 2 seconds before displaying anything else
  }

  Serial.println("POST completed");

  // clear timers used to assess system slack time
//  for (int zT = 0;zT < 4; zT++) {t0[zT] = micros(); t2[zT] = 0;}  // clear the timings
}

// ---------------------------------------------------------------------------------

void setDefaults() {
  // load default values
  accelerometer_data_rawX_i = 1;  // instantaneous record for averaging
  accelerometer_data_rawY_i = 2;  // instantaneous record for averaging
  autoOFF = 0;            // if set > 0 then counts down to servo switch OFF, if -1 disabled
  autoRest = autoRestMax; // returns to a rest position after a timeout
  Blink = false;          // LED blink clock
  BlinkCnt = 0;           // counter for LED Blink clock
  BlockAsk = 3;           // number of blocks asked for from the Pixy2 camera, default = 3
  BlockCnt = 0;           // latest number of blocks detected by the Pixy2 camera
  Border = false;         // if true then display has a border
  Bright = 255;           // FastLED brightness
  Brightness = 255;       // Display brightness
  CamerLux = 50;          // camera brightness, default 50
  C_Cnt = 0;              // counter used in 'C' button detection
  C_Dn = 0;               // counter for 'C' button down period
  cmdMode = ' ';          // command mode
  cmdSgn = 1;             // cmdVal sign, 1 or -1
  cmdType = ' ';          // command mode type
  cmdVal = 0;             // value associated with a cmdType
  CZ = 0;                 // received button values
  CZ_Wait = false;        // flag used for button release wait function
  DIR_FL = 1;             // direction sign, +ve forward, -ve backwards
  DIR_FR = 1;             // direction sign, +ve forward, -ve backwards
  DIR_RL = 1;             // direction sign, +ve forward, -ve backwards
  DIR_RR = 1;             // direction sign, +ve forward, -ve backwards
  DispBack = false;       // flag used to draw a back arrow on the display
  DispChng = false;       // flag invokes initialisation code
  DispClr = false;        // set == true to clear the memory contents
  DispCnt = 0;            // counter used in display updating
  DispDel = 0;            // >0 display delay counter, stops updates over I2C
  DispDisp = false;       // set = true to immediately display memory contents
  DispMenu = 0;           // group type, 0 = sensors, 1 = tasks, etc
  DispMode = 0;           // the type of display for a given mode
  DispModeCnt = 0;        // counter used in display updating
  DispMon = false;        // == true if connected to OLED Mirror app
  DispNOP = false;        // set == true to stop display I2C updates
  DispNow = 0;            // if > 0 display special modes
  DispNxtCnt = 0;         // counter used in auto-branching of display
  DispPnt = 0;            // points at the current display mode, default = 0 batteries
  DispTask = 0;           // task pointer for display functions
  DispTx = 0;             // counter used with display mirror app
  DispVal = 0;            // general display value, passed from tasks
  Dither = 25;            // overlay signal used to reduce stiction
  Drive = 0;              // if > 0 then driving in various directions
  DriveLast = 0;          // previous driving direction
  Gear = 1;               // value used to determine PWM Max values
  ESC = false;            // set = true to escape all functions
  HeadActive = false;     // true if head is in scan mode for tasks
  HeadAtt = false;        // = true when head servo has been attached
  HeadDir = 0;            // head movement direction, +1 == RH, -1 == LH
  HeadInterval = 20000;   // main loop head movement interval in microseconds
  HeadPause = 0;          // 20ms counter to temporarily pause head movement
  HeadSubTask = 0;        // head movement sub task pointer
  HeadTask = 0;           // head movement task pointer
  HeadTaskRun = false;    // set to true during head tasks
  HeadTgtCnt = 0;         // counter used in automatic head servo move to target actions
  HeadVal = Head_90 + 200; // current head servo angle, set to 200 > Head_90
  HeadTgt = HeadVal;      // target head servo angle
  I2C_Disp = false;       // true if display is detected during POST
  I2C_LTOF = false;       // true if VL53L1X is detected during POST
  I2C_MPU = false;        // true if MPU6050 is detected during POST
  I2C_Pixy = false;       // true if Pixy2 is detected during POST
  Int_En = false;         // interrupts enabled flag, == true if enabled
  IntMask = B0000;        // 4-bit mask used to flag the occurence of interrupts
  JoyActive = true;       // flag used to prevent stack overflow during moves
  JoyDL = 0;              // direction 1 counter
  JoyMode = 0;            // direction of travel flag, 0 = stationary, 1 = forward, -1 = backward
  JoySpd = 0;             // speed vector = max(JoyX,JoyY)
  JoyX = 0;               // received value, default
  JoyXV = 0;              // converted value 0 - 127
  JoyY = 0;               // received value, default
  JoyYi = 0;              // tracking filter for received JoyY value
  JoyYV = 0;              // converted value 0 - 128
  JX = 0;                 // global variable used in turning, with Joystick app
  JY = 0;                 // global variable used in turning, with Joystick app
  LED_Cnt = 0;            // counter used in LED sequencing
  LED_Del = 0;            // delay timer used in LED functions
  LED_Period = 20;        // LED timer period, default = 20ms
  LED_Run = false;        // == true if LED_Task is to run
  LEDshow = false;        // LED show flag for mouth LEDs
  LedSpd = PWM_StartMin;  // motor speed value used by LED animation
  LED_Task = 0;           // LED task pointer
  LedTick = false;        // set true during LED cycles
  LEDVal = 0;             // values used in LED tasks
  Lights = false;         // if == true then turn Pixy2 lights ON
  LTOFcalc = false;       // if == true then calculate speed of ranged object in m/s
  LTOFspd10 = 0;          // speed of ranged target in m/s x 10
  LTOFspd10L = 0;         // previous speed of ranged target in m/s x 10
  LTOFspdCnt = 0;         // speed peak detector sustain limit
  MainMode = 0;           // PixyBot main state, 0 = sleep, 1 = Drive, 2 = LTOF, 3 = Pixy2
  MainTask = 0;           // main task pointer
  MainTaskNext = 0;       // the task to run after an arrest; default is 0
  ModeSel = 0;            // MainMode selection
  MoveState = 0;          // move state machine pointer
  period16ms = 16;        // a period which varies between 16-17ms to acjhieve 60Hz
  phase16ms = 0;          // a 3-phase pointer which adjusts the 16ms value to get 16.6666
  Ping = 0;               // 'ping' counter
  PixyBotActive = false;  // = true if PixyBot is active, otherwise false
  PixyDel = 0;            // delay timer used in checking for Pixy2
  PixyEn = false;         // == true when I2C to Pixy2 is active
  PixyFps = 60;           // camera frames per second, indicates light level quality
  PwmBckMax = 128;        // max reverse PWM used in LTOF demo
  PwmEn = false;          // if true motor PWM is enabled
  PWM_FL = 0;             // PWM applied to motor drive +/-255
  PWM_FR = 0;             // PWM applied to motor drive +/-255
  PwmFwdMax = 128;        // max forward PWM used in LTOF demo
  PwmInt = 0;             // integrator used in motor control
  PWM_RL = 0;             // PWM applied to motor drive +/-255
  PWM_RR = 0;             // PWM applied to motor drive +/-255
  RangeEn = false;        // == true if VL53L1X ranging is active
  RangeMax = LtofLimit;   // the current limited to Range measurements
  RangeMin = LtofMin;     // the current lower limited to Range measurements
  readWiiCall = false;    // flag prevents call stack overflow
  Rx_Chk = 0;             // Rx checksum
  RxCZ = 0;               // buffered received CZ value
  RxJoyX = 0;             // buffered received JoyX value
  RxJoyY = 0;             // buffered received JoyY value
  Rx_len = 0;             // length of WiFi received data block
  RxRec = false ;         // set true when a valid frame of data is received
  RxState = 0;            // receiver state machine state
  Rx_Task = 0;            // pointer user by Rx state machine
  RTxTimeout = 0;         // counter used to auto-reset if WiFi fails
  RxVal = -1;             // value received from serial Rx
  SerialRx = false;       // == true if any character has been received
  ServoInc = 1;           // servo test increment direction
  ServoPnt = 0;           // servo target pointer for manual control, default = 0
  ServoVal = 500;         // value sent to both Pixy2 servos over I2C, default = 500
  SetSel = 0;             // setting selection pointer
  SigAp = 0;              // block array pointer to SigId data
  SigHi = 250;            // object signtaure ID width trigger point, large
  SigId = 0;              // object of interest signtaure ID, default is 0 (largest)
  SigLo = 200;            // object signtaure ID width trigger point, small
  SigTask = 0;            // object of interest recognition task pointer
  SlotEn = false;         // if == true then constantly monitor wheel slot encoders
  Snore = 3;              // snore animation character
  SubCnt = 0;             // sub task general counter
  SubDel = 0;             // subtask delay counter
  SubNext = 0;            // pointer to next sub task
  SubTask = 0;            // sub task pointer
  sw0Cnt = 0;             // button switch counter
  sw0DwnTime = 0;         // button pressed down time
  sw0LastState = HIGH;    // previous state of button switch, HIGH/LOW
  sw0_Nop = false;        // if == true don't perform switch functions
  sw0State = HIGH;        // state of read button switch pin
  sw0Timer = 0;           // timer used to detemine button sequences
  sw0Wup = false;         // set == true for SW0 to wait for button release
  sw1Cnt = 0;             // button switch counter
  sw1LastState = HIGH;    // previous state of button switch, HIGH/LOW
  sw1_Nop = false;        // if == true don't perform switch functions
  sw1State = HIGH;        // state of read button switch pin
  sw1Timer = 0;           // timer used to detemine button sequences
  sw1Wup = false;         // set == true for SW1 to wait for button release
  Task16ms = 0;           // 16.666ms subtask timer
  Task40ms = 0;           // 40ms subtask timer
  Tx_PingCnt = 0;         // 1 second Tx ping timer
  USB = false;            // == true if initial battery is below 5.5v
  VL53L1X_OC = 199;       // VL53L1X ROI SPAD centre
  VL53L1X_ROIX = 16;      // VL53L1X ROI SPAD width
  VL53L1X_ROIY = 16;      // VL53L1X ROI SPAD height
  VL53L1X_SkipDel = 31;   // read function skip timer delay tracker, nominally 32
  VL53L1X_Task = 0;       // if > 0 then run LV53L1X tasks
  WiFiCntC = 0;           // C button counter for WiFi enable
  WiFiConCnt = 0;         // counter used in connection retry times
  WiFiConnected = false;  // == true if a WiFi connection has been established
  WiFiEn = false;         // true when WiFi is enabled with prolonged 'Z' button activity
  WiFiPing = 0;           // Rx counter used to maintain connected state
  WiFiRx = false;         // true when a block of data has been received
  WiFiRxBytes = 0;        // number of bytes receiv ed over the WiFi link
  WiFiRxErr;              // number of receive errors in the WiFi link
  WiFiRxRec = false;      // true when a block of data has been successfully received
  WiFiTryCnt = 0;         // WiFi try to connect count
  WiFiTryNum = 0;         // WiFi try to connect total count
  WiFiTx = false;         // true when a block of data has been sent successfully
  WiFiTxBytes = 0;        // number of bytes receiv ed over the WiFi link
  WiFiTx_CB = 1;          // == 1 call back is clear to send, == -1 resend
  WiFiTx_Chk = 0;         // Tx checksum
  WiFiTx_len = 0;         // >0 when WiFi Tx buffer contains data to be sent
  WiFiTxErr = 0;          // WiFi data error
  WiFiTxErrCnt = 0;       // WiFi Tx error count, leads to disconnection
  WiiType = 0;            // device type, 0 == Nunchuk, 1 == Classic
  Z_Dn = 0;               // button pressed state; 0 = UP, 1 = Down
  Z_Mode = false;         // joystick motion flag, normally false, if == true then slide mode
}

// ---------------------------------------------------------------------------------

void synchLoopTimers(unsigned long zAdvance) {
  // reset or adjust main loop timers, based on zAdvance
  // if  0 then a 2ms phase offset is introduced to prevent simultaneous triggering
  // if >0 then add zAdvance to the current timer values
  if (zAdvance == 0) {
    // setting up fresh loop timers
    LED_Timer = millis() + 26;  // reset 20+ms loop timer with 3ms phase to 10ms timer
    loopDelay = millis();
    next16ms = millis() + 2;    // reset 16.666ms task timer with 2ms phase to loopDelay
    next40ms = millis() + 4;    // reset 40ms task timer with 4ms phase to loopDelay
    print40ms = millis() + 46;  // reset 40ms printer loop timer with 4ms phase to loopDelay
  } else {
    // advancing loop timers
    LED_Timer += zAdvance;
    loopDelay += zAdvance;
    next16ms += zAdvance;
    next40ms += zAdvance;
    print40ms += zAdvance;
  }
}

// --------------------------------------------------------------------------------

/*
  Timing analysis used to determine which functions are taking the longest. This is
  done using the t0[],t1[] and t2[] arrays, allowing you to place up to 10 measurements
  in each compiled version.

  System          20    µs
  ReadWiFiRx      76    µs
  Print40ms       321   µs
  FastLED.show()  542   µs
  Display_40ms()  4513  µs
  Task_16ms()     24636 µs
  Loop()          25052 µs

  Conclusion, these delays impact on slot reading, so move that task to Core 0
  
*/

// --------------------------------------------------------------------------------
